import sys
import logging
import argparse

from auto_portforward.tui import ProcessMonitor

LOGGER = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description=(
            "Process monitoring and port forwarding tool. Monitors processes with open ports "
            "and allows toggling port forwarding for individual processes or groups."
            "When monitoring a remote SSH host, ports can be automatically forwarded. "
            "When targeting a process group, port forwarding is automatically managed as "
            "processes enter or leave the group - new processes in the group will have "
            "their ports forwarded, and forwarding is removed when processes exit."
        )
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-l", "--local", action="store_true", help="Use local process monitor")
    group.add_argument("--mock", action="store_true", help="Use mock process monitor")
    parser.add_argument(
        "ssh_host",
        nargs="?",
        help="SSH host to connect to (for remote process monitor and portforwarding)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable verbose logging",
    )
    parser.add_argument(
        "--log-file",
        help="Path to log file",
    )
    args = parser.parse_args()

    if args.verbose:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)

    if args.log_file:
        file_handler = logging.FileHandler(args.log_file)
        file_handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
        root_logger = logging.getLogger()
        root_logger.addHandler(file_handler)

    if args.local:
        from auto_portforward.process_provider.local import LocalProcessMonitor

        monitor = LocalProcessMonitor()
    elif args.mock:
        from auto_portforward.process_provider.local import MockProcessMonitor

        monitor = MockProcessMonitor()
    else:
        if not args.ssh_host:
            parser.error("SSH host is required when not using local or mock process monitor")
            sys.exit(1)

        from auto_portforward.process_provider.ssh_remote import RemoteProcessMonitor

        monitor = RemoteProcessMonitor(args.ssh_host)
        if not monitor.connect():
            LOGGER.error("Failed to establish connection. Exiting.")
            sys.exit(1)

    app = ProcessMonitor(monitor)
    app.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
This is a remote process monitor that sends process and connection information to a local socket.
It is used to monitor processes on a remote machine.

To be run on the remote machine.
"""

import socket
import json
import time
import sys

from dataclasses import asdict

# if we are in ssh_single_file_mode
# we directly inject the get_connections and get_processes functions
# into the local namespace
if not locals().get("ssh_single_file_mode", False):
    from .get_process_with_openports import get_connections, get_processes


def send_via_socket():
    """
    This is a script that is run on the remote machine.
    It sends process and connection information to a local socket.
    It is used to monitor processes on a remote machine.
    To be run on the remote machine.
    """
    if len(sys.argv) != 2:
        print("Usage: python3 remote_monitor.py <port>")
        sys.exit(1)

    port = int(sys.argv[1])
    print(f"Connecting to local socket on port {port}")
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(("localhost", port))
    print("Connected to local socket")

    while True:
        try:
            # Send process and connection information
            connections, udp_connections = get_connections()
            data = {
                "type": "data",
                "processes": {str(k): asdict(v) for k, v in get_processes(connections, udp_connections).items()},
            }
            msg = json.dumps(data).encode()
            length_bytes = len(msg).to_bytes(4, "big")
            print(f"Sending data message, length: {len(msg)}")
            s.sendall(length_bytes + msg)
            time.sleep(1.5)  # Update every second
        except Exception as e:
            import traceback

            print(f"Error in main loop: {e}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            break

    print("Closing connection")
    s.close()


if __name__ == "__main__":
    send_via_socket()



from auto_portforward import datatype
from . import abstract_provider
from . import get_process_with_openports


class MockProcessMonitor(abstract_provider.AbstractProvider):
    def __init__(self):
        super().__init__()
        self.processes: dict[int, datatype.Process] = {}

    async def get_processes(self) -> dict[str, datatype.Process]:
        # Create some mock processes with listening ports
        mock_processes = {
            "1234": datatype.Process(
                pid=1234,
                name="nginx",
                cwd="/etc/nginx",
                status="running",
                create_time="1234567890",
                tcp=[80, 443],
            ),
            "5678": datatype.Process(
                pid=5678,
                name="python",
                cwd="/home/user/code",
                status="running",
                create_time="1234567891",
                tcp=[8000],
            ),
            "5679": datatype.Process(
                pid=5679,
                name="python",
                cwd="/home/user/code",
                status="running",
                create_time="1234567893",
                tcp=[8005],
            ),
            "9012": datatype.Process(
                pid=9012,
                name="postgres",
                cwd="/var/lib/postgresql",
                status="running",
                create_time="1234567892",
                tcp=[5432],
            ),
            "9013": datatype.Process(
                pid=9013,
                name="dns",
                cwd="/etc/bind",
                status="running",
                create_time="1234567893",
                tcp=[],
                udp=[53],
            ),
        }
        return mock_processes


class LocalProcessMonitor(abstract_provider.AbstractProvider):
    def __init__(self):
        super().__init__()
        self.processes: dict[int, datatype.Process] = {}

    async def get_processes(self) -> dict[str, datatype.Process]:
        connections, udp_connections = get_process_with_openports.get_connections()
        processes = get_process_with_openports.get_processes(connections, udp_connections)
        self.processes = processes
        return {str(k): v for k, v in self.processes.items()}



import os
import socket
import sys
import logging

LOGGER = logging.getLogger(__name__)

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    import os
    import subprocess

    HAS_PSUTIL = False
    LOGGER.info("psutil not found, using fallback methods")

# If we are in ssh_single_file_mode
# we directly inject the Process class
# into the local namespace
if not locals().get("ssh_single_file_mode", False):
    from ..datatype import Process


def get_cwd_linux(pid: int) -> str:
    try:
        password = os.getenv("AP_SUDO_PASSWORD")
        if password:
            cmd = ["sudo", "-S", "readlink", f"/proc/{pid}/cwd"]
            result = subprocess.run(cmd, input=password + "\n", capture_output=True, text=True)
            if result.returncode == 0:
                return result.stdout.strip()
        else:
            return os.readlink(f"/proc/{pid}/cwd")
    except Exception:
        pass
    return "?"


def get_cwd_macos(pid: int) -> str:
    try:
        result = subprocess.check_output(["lsof", "-a", "-p", str(pid), "-d", "cwd"], text=True)
        lines = result.strip().split("\n")
        if len(lines) >= 2:
            # Last column of the second line is the cwd
            return lines[1].split()[-1]
    except Exception:
        pass
    return "?"


def get_cwd_fallback(pid: int) -> str:
    if sys.platform.startswith("linux"):
        return get_cwd_linux(pid)
    elif sys.platform.startswith("darwin"):
        return get_cwd_macos(pid)
    else:
        return "?"


def get_processes(connections: dict[int, list[int]], udp_connections: dict[int, list[int]]) -> dict[int, Process]:
    processes = {}

    for pid in connections.keys() | udp_connections.keys():
        try:
            pid = int(pid)
        except (ValueError, TypeError):
            continue

        if HAS_PSUTIL:
            proc = psutil.Process(pid)
            p = Process(
                pid=proc.pid,
                name=proc.name(),
                cwd=proc.cwd(),
                status=proc.status(),
                create_time=str(proc.create_time()),
                tcp=sorted(connections.get(pid, [])),
                udp=sorted(udp_connections.get(pid, [])),
            )
            processes[p.pid] = p

        else:
            # Fallback: basic info using single ps command
            # Get process name, status, and start time (no cwd available)
            cmd = ["ps", "-p", str(pid), "-o", "comm=,stat=,lstart="]
            output = subprocess.check_output(cmd, text=True).strip()
            if not output:
                continue

            parts = output.split()
            name = parts[0]
            status = parts[1]
            create_time = " ".join(parts[2:])

            p = Process(
                pid=pid,
                name=name,
                cwd=get_cwd_fallback(pid),
                status=status,
                create_time=create_time,
                tcp=sorted(connections.get(pid, [])),
                udp=sorted(udp_connections.get(pid, [])),
            )
            processes[p.pid] = p

    return processes


def get_connections(sudo_password: str | None = None) -> tuple[dict[int, list[int]], dict[int, list[int]]]:
    """
    Get a mapping of process IDs to listening ports for both TCP and UDP.
    If needs_sudo is True and sudo_password is provided, use sudo -S and pass the password via stdin.
    Returns a tuple of (tcp_connections, udp_connections)
    """
    sudo_password = sudo_password or os.getenv("AP_SUDO_PASSWORD")

    tcp_connections: dict[int, set[int]] = {}
    udp_connections: dict[int, set[int]] = {}

    def mapper(connections: dict[int, set[int]]) -> dict[int, list[int]]:
        return {k: list(v) for k, v in connections.items()}

    if HAS_PSUTIL:
        for c in psutil.net_connections():
            if c.status == "LISTEN":
                if c.type == socket.SOCK_STREAM:
                    container = tcp_connections.setdefault(c.pid, set())
                    container.add(c.laddr[1])
                elif c.type == socket.SOCK_DGRAM:
                    container = udp_connections.setdefault(c.pid, set())
                    container.add(c.laddr[1])
        return mapper(tcp_connections), mapper(udp_connections)
    else:
        # Fallback using 'lsof' (Unix only)
        try:
            args = []
            if sudo_password:
                args = ["sudo", "-S"]
            args += [
                "lsof",
                "-nP",
                "-iTCP",
                "-iUDP",
            ]
            try:
                if sudo_password is not None:
                    LOGGER.debug("Running lsof with sudo")
                    output = subprocess.check_output(
                        args, text=True, input=sudo_password + "\n", stderr=subprocess.PIPE
                    )
                else:
                    LOGGER.debug("Running lsof without sudo")
                    output = subprocess.check_output(args, text=True, stderr=subprocess.PIPE)
            except subprocess.CalledProcessError as e:
                LOGGER.error("Failed to run lsof command: %s", e)
                LOGGER.error("stdout: %s", e.output)
                LOGGER.error("stderr: %s", e.stderr)
                raise
            for line in output.splitlines()[1:]:
                parts = line.split()
                if len(parts) < 9:
                    continue
                pid = int(parts[1])
                port_info = parts[8]
                if ":" not in port_info:
                    continue
                port_str = port_info.rsplit(":", 1)[-1]
                if not port_str.isdigit():
                    continue
                port = int(port_str)
                proto = parts[7]
                if proto.startswith("TCP") and "LISTEN" in line:
                    container = tcp_connections.setdefault(pid, set())
                    container.add(port)
                elif proto.startswith("UDP"):
                    container = udp_connections.setdefault(pid, set())
                    container.add(port)
        except Exception as e:
            LOGGER.error(f"exception: {e}")
            raise e

    return mapper(tcp_connections), mapper(udp_connections)


if __name__ == "__main__":
    connections, udp_connections = get_connections()
    print("tcp connections", connections)
    print("udp connections", udp_connections)
    processes = get_processes(connections, udp_connections)
    print("processes", processes)



#!/usr/bin/python
import asyncio
import logging
import os
import threading
from typing import Dict, Set
from textual import on, work
from textual.app import App, ComposeResult
from textual.message_pump import Timer
from textual.message import Message
from textual.widgets import Header, Footer, Static, Tree
from textual.binding import Binding
from textual.widgets import Log


# from textual.style import Style
from rich.style import Style
from rich.text import Text

from auto_portforward.process_provider.abstract_provider import AbstractProvider

from .process_provider.ssh_remote import RemoteProcessMonitor
from .datatype import Process

# Configure logging
log_dir = os.path.expanduser("logs")
os.makedirs(log_dir, exist_ok=True)
log_file = os.path.join(log_dir, "tui.log")

# Create formatters
FORMATTER = logging.Formatter("[%(asctime)s %(levelname)-5s] %(message)s", datefmt="%H:%M:%S")

# Configure root logger
root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)


LOGGER = logging.getLogger(__file__)

GROUP_SELECTED_STYLE = Style(color="green", italic=True)
NODE_SELECTED_STYLE = Style(color="yellow", italic=True)


class ProcessTree(Tree):
    """
    A tree of processes.
    """

    def __init__(self, monitor: AbstractProvider, logger: Log):
        super().__init__(monitor.name)
        self.monitor: AbstractProvider = monitor
        self.last_memory: Dict[str, Process] = {}
        self.selected_groups: Set[str] = set()
        self.selected_processes: Set[int] = set()
        self.group_by = "cwd"
        self.sort_reverse = False
        self.filter_text = ""
        self.update_interval = 1.0
        self.last_update = 0
        self.regular_update_timer: Timer | None = None
        self.logger: Log = logger

    def on_mount(self) -> None:
        def expand_all(node):
            node.expand()
            for child in node.children:
                expand_all(child)

        expand_all(self.root)
        # Start continuous update loop
        # self.regular_update_timer = self.set_interval(
        #     lambda: self.update_processes,
        #     self.update_interval,
        # )

        if not self.call_later(self.update_processes):
            raise RuntimeError("Failed to schedule update_processes")

    def is_new_memory(self, new_memory: Dict[str, Process]) -> bool:
        if not self.last_memory:
            return True
        if len(new_memory) != len(self.last_memory):
            return True
        for pid, process in new_memory.items():
            if process != self.last_memory[pid]:
                return True
        return False

    @work(exclusive=True)
    async def update_processes(self) -> None:
        new_memory = await self.monitor.get_processes()
        if new_memory and self.is_new_memory(new_memory):
            self.last_memory = new_memory.copy()
            await self.update_process_layout()

        await asyncio.sleep(1)
        # Schedule next update immediately
        self.call_later(self.update_processes)

    async def toggle_group(self, group_key: str) -> None:
        LOGGER.debug("Toggling group: %s", group_key)
        # Convert Text object to string if needed
        if hasattr(group_key, "plain"):
            group_key = group_key.plain
        if group_key in self.selected_groups:
            self.selected_groups.remove(group_key)
        else:
            self.selected_groups.add(group_key)
        await self.update_process_layout()

    async def toggle_process(self, pid: int) -> None:
        if pid in self.selected_processes:
            self.selected_processes.remove(pid)
        else:
            self.selected_processes.add(pid)
        await self.update_process_layout()

    async def update_process_layout(self) -> None:
        ports_to_forward = set()

        # Group processes
        grouped: Dict[str, list[Process]] = {}
        for pid, process in self.last_memory.items():
            if self.filter_text and self.filter_text.lower() not in process.name.lower():
                continue
            key = getattr(process, self.group_by)
            if key not in grouped:
                grouped[key] = []
            grouped[key].append(process)

        # Sort groups
        sorted_groups = sorted(
            grouped.items(),
            key=lambda x: str(x[0]) if x[0] is not None else "",
            reverse=self.sort_reverse,
        )

        # clear any existing nodes
        self.clear()

        # Create tree structure
        for group, processes in sorted_groups:
            group_key = str(group) if group is not None else "Unknown"
            selected_by_group = group_key in self.selected_groups

            if self.group_by != "pid":
                # Create group node
                group_node = self.root.add(group_key, expand=True)
                group_node.data = {"is_group": True}

                if selected_by_group:
                    group_node.label.style = GROUP_SELECTED_STYLE

                group_or_root_node = group_node
            else:
                # PID does not needs grouping.
                group_or_root_node = self.root

            # Sort processes
            sorted_processes = sorted(
                processes,
                key=lambda p: str(p.pid) if p.pid is not None else "0",
                reverse=self.sort_reverse,
            )

            for process in sorted_processes:
                parts = []
                if process.tcp:
                    parts.extend(
                        [
                            (" 🌐", "bold cyan"),
                            ("TCP", "bold cyan u"),
                            (": ", "bold cyan"),
                            f"{','.join(map(str, process.tcp))}",
                        ]
                    )
                if process.udp:
                    parts.append((" 📡UDP: ", "bold red"))
                    parts.append(f"{','.join(map(str, process.udp))}")

                process_node = group_or_root_node.add_leaf(
                    Text.assemble(
                        ("🆔", ""),
                        (f"{process.pid}", "bold"),
                        (" 📦", ""),
                        (f"{process.name}", "blue"),
                        *parts,
                        (f" (⚡{process.status})", ""),
                        overflow="ellipsis",
                        justify="center",
                    )
                )
                # Add process node
                process_node.data = {"is_group": False, "pid": process.pid}

                # selected can also be done on a node-level
                if selected_by_group:
                    process_node.label.style = GROUP_SELECTED_STYLE
                elif process.pid in self.selected_processes:
                    process_node.label.style = NODE_SELECTED_STYLE
                else:
                    continue

                # Add ports to forward
                for p in process.tcp:
                    ports_to_forward.add(p)

        self.call_later(self.update_toggled_ports, ports_to_forward)

    @work(exclusive=True)
    async def update_toggled_ports(self, ports_to_forward: Set[int]) -> None:
        await self.monitor.set_toggled_ports(ports_to_forward)

    async def on_unmount(self) -> None:
        """Clean up all port forwarding processes when the widget is removed."""

        if self.regular_update_timer:
            self.regular_update_timer.stop()

        await self.monitor.cleanup()

    async def change_group_by(self) -> None:
        options = ["cwd", "name", "pid"]
        current_index = options.index(self.group_by)
        self.group_by = options[(current_index + 1) % len(options)]
        await self.update_process_layout()

    async def toggle_sort(self) -> None:
        self.sort_reverse = not self.sort_reverse
        await self.update_process_layout()

    async def set_filter(self, text: str) -> None:
        self.filter_text = text
        await self.update_process_layout()


class TuiLogHandler(logging.Handler):
    class NewLog(Message):
        """
        This is a message that is sent to the TUI logger.
        """

        def __init__(self, msg: str):
            super().__init__()
            self.msg = msg

    def __init__(self, tui_logger: Log):
        super().__init__()
        self.tui_logger = tui_logger
        self._lock = threading.Lock()

    def emit(self, record):
        msg = self.format(record)
        try:
            self.tui_logger.post_message(self.NewLog(msg))
        except Exception:
            self.handleError(record)


class ProcessMonitor(App):
    CSS = """
    Tree > .selected-group {
        color: yellow;
        text-style: bold;
    }
    Tree > .selected-process {
        color: yellow;
    }
    #log-widget {
        height: 20%;
    }
    """

    BINDINGS = [
        Binding("g", "change_group_by", "Change Group By"),
        Binding("s", "toggle_sort", "Toggle Sort"),
        Binding("f", "filter", "Filter"),
        Binding("t", "toggle_group", "Toggle Group"),
        Binding("q", "quit", "Quit"),
    ]

    def __init__(self, monitor: RemoteProcessMonitor):
        super().__init__()
        self.monitor = monitor
        self.logger = Log(id="log-widget", max_lines=50)
        self.process_tree = ProcessTree(monitor, self.logger)

    @on(TuiLogHandler.NewLog)
    def handle_new_log(self, message: TuiLogHandler.NewLog) -> None:
        """
        These messages are bubbled up from the TUILogHandler.
        """
        self.logger.write_line(message.msg)

    def on_mount(self) -> None:
        # Attach TUI log handler
        tui_log_handler = TuiLogHandler(self.logger)
        tui_log_handler.setLevel(logging.DEBUG)
        tui_log_handler.setFormatter(FORMATTER)

        # Add handler to root logger to capture all logs
        root_logger = logging.getLogger()
        root_logger.addHandler(tui_log_handler)
        self.post_message(TuiLogHandler.NewLog("[Log Area]"))

    def compose(self) -> ComposeResult:
        yield Header()
        yield self.process_tree
        yield self.logger
        yield Footer()

    async def action_change_group_by(self) -> None:
        await self.process_tree.change_group_by()

    async def action_toggle_sort(self) -> None:
        await self.monitor.cleanup()
        await self.process_tree.toggle_sort()

    def action_filter(self) -> None:
        self.app.push_screen(FilterScreen(self.process_tree))

    async def action_toggle_group(self) -> None:
        node = self.process_tree.cursor_node
        if node.is_root:
            return
        try:
            is_group = node.data.get("is_group")
        except Exception:
            is_group = False
        if is_group:
            # Convert Text object to string if needed
            label = node.label.plain if hasattr(node.label, "plain") else str(node.label)
            await self.process_tree.toggle_group(label)
        else:
            # If it's a process node, toggle its parent group
            pid = node.data.get("pid", None)
            if pid:
                await self.process_tree.toggle_process(pid)
            # parent = node.parent
            # if parent:
            #     # Convert Text object to string if needed
            #     label = parent.label.plain if hasattr(parent.label, "plain") else str(parent.label)
            #     await self.process_tree.toggle_group(label)

    async def on_unmount(self) -> None:
        """Clean up resources when the app is closed."""
        await self.monitor.cleanup()


class FilterScreen(App):
    def __init__(self, process_tree: ProcessTree):
        super().__init__()
        self.process_tree = process_tree

    def compose(self) -> ComposeResult:
        yield Static("Enter filter text:")
        yield Static(self.process_tree.filter_text)

    async def on_key(self, event):
        if event.key == "escape":
            self.app.pop_screen()
        elif event.key == "enter":
            await self.process_tree.set_filter(self.process_tree.filter_text)
            self.app.pop_screen()
        else:
            self.process_tree.filter_text += event.character



import logging

from abc import ABC, abstractmethod
from typing import Set


from auto_portforward import datatype

LOGGER = logging.getLogger(__file__)


class AbstractProvider(ABC):
    def __init__(self):
        self.toggled_ports: Set[int] = set()

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    async def get_processes(self) -> dict[str, datatype.Process]:
        pass

    async def cleanup(self) -> None:
        for port in self.toggled_ports:
            await self.on_ports_turned_off(port)
        self.toggled_ports.clear()

    async def on_ports_turned_on(self, port: int):
        LOGGER.info("Port %i is turned on", port)

    async def on_ports_turned_off(self, port: int):
        LOGGER.info("Port %i is turned off", port)

    async def set_toggled_ports(self, ports: Set[int]) -> None:
        """
        This method is used to just manage ports on-off event.
        """

        # Remove old port forwards
        for existing in list(self.toggled_ports):
            if existing not in ports:
                await self.on_ports_turned_off(existing)
                self.toggled_ports.remove(existing)
        # Start new port forwards
        for p in ports:
            if p not in list(self.toggled_ports):
                await self.on_ports_turned_on(p)
                self.toggled_ports.add(p)



import os
import signal
import sys
import ctypes


def set_pdeathsig(sig=signal.SIGTERM):
    """Set parent death signal on Linux so child dies if parent dies."""
    if sys.platform.startswith("linux"):
        libc = ctypes.CDLL("libc.so.6")
        return libc.prctl(1, sig)
    return 0


def preexec_set_pdeathsig():
    # Set the process group ID to the current process ID.
    os.setpgrp()
    # Set the session ID to the current process ID.
    # the following is done via the popen call
    # os.setsid()
    # Set the parent death signal to the current process ID.
    set_pdeathsig(signal.SIGTERM)



import subprocess
import logging

import os
import signal
import atexit

from auto_portforward.utils import preexec_set_pdeathsig

LOGGER = logging.getLogger(__file__)


class SSHForward:
    def __init__(self, port: int, ssh_host: str):
        # port to forward
        self.port = port
        # ssh host to connect to
        self.ssh_host = ssh_host
        # whether the cleanup has been called
        self.had_cleanup = False
        # the process that is running the port forwarding
        self.process: subprocess.Popen | None = None
        # register cleanup to be called when the program exits
        atexit.register(self.cleanup)

    def start(self):
        self.process = subprocess.Popen(
            ["ssh", "-N", "-L", f"{self.port}:localhost:{self.port}", self.ssh_host],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=preexec_set_pdeathsig,
        )
        LOGGER.info(
            "Started port forwarding for port %s with PID %s",
            self.port,
            self.process.pid,
        )

    def cleanup(self):
        if self.had_cleanup:
            return
        try:
            self.process.terminate()
            os.kill(self.process.pid, signal.SIGINT)
            # # Kill the entire process group
            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
            self.process.wait(timeout=5)
            LOGGER.info("Terminated port forwarding for port %s", self.port)
        except subprocess.TimeoutExpired:
            LOGGER.warning(
                "Port forwarding process for port %s did not terminate gracefully, forcing...",
                self.port,
            )
            os.kill(self.process.pid, signal.SIGKILL)
            os.killpg(os.getpgid(self.process.pid), signal.SIGKILL)
        except Exception as e:
            LOGGER.error("Error terminating port forwarding for port %s: %s", self.port, e)
        self.had_cleanup = True



from dataclasses import dataclass, field


@dataclass
class Process:
    pid: int
    name: str
    cwd: str
    status: str
    create_time: str
    tcp: list[int] = field(default_factory=list)
    udp: list[int] = field(default_factory=list)



import json
import logging
import os
import socket
import subprocess
import threading
import signal

from pathlib import Path
from dataclasses import dataclass, field

from auto_portforward.ssh_port_forward import SSHForward
from auto_portforward.utils import preexec_set_pdeathsig
from .abstract_provider import AbstractProvider
from . import get_process_with_openports, script_on_remote_machine
from .. import ROOT_DIR, datatype

THIS_DIR = Path(__file__).parent

LOGGER = logging.getLogger(__name__)


def build_ssh_single_file_mode_script():
    remote_script = 'locals()["ssh_single_file_mode"] = True\n'
    with open(ROOT_DIR / datatype.__file__, "r") as f:
        remote_script += f.read() + "\n"
    with open(THIS_DIR / get_process_with_openports.__file__, "r") as f:
        remote_script += f.read() + "\n"
    with open(THIS_DIR / script_on_remote_machine.__file__, "r") as f:
        remote_script += f.read() + "\n"

    return remote_script


@dataclass
class SharedMemory:
    processes: dict[str, datatype.Process]
    lock: threading.Lock = field(default_factory=threading.Lock)
    has_new_data: threading.Event = field(default_factory=threading.Event)
    is_finished: threading.Event = field(default_factory=threading.Event)


def run_remote_script(ssh_host: str, shared_memory: SharedMemory, monitor_instance: "RemoteProcessMonitor"):
    import time

    # time.sleep(2)
    try:
        # Create a local socket for communication
        LOGGER.debug("Creating local socket")
        local_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        local_socket.bind(("localhost", 0))  # Bind to localhost
        local_socket.listen(1)
        port = local_socket.getsockname()[1]
        LOGGER.debug("Created local socket on port %d", port)

        # Read the remote script
        LOGGER.debug("Reading remote script from: %s", THIS_DIR)

        # Start the remote Python process that will connect back to us
        remote_cmd = f"python3 -c '{build_ssh_single_file_mode_script()}' {port}"
        LOGGER.debug("Starting SSH process with port forwarding")
        ssh_process = subprocess.Popen(
            [
                "ssh",
                "-R",
                f"{port}:localhost:{port}",
                ssh_host,
                f"AP_SUDO_PASSWORD={os.getenv('AP_SUDO_PASSWORD', '')} {remote_cmd}",
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,
            universal_newlines=True,
            preexec_fn=preexec_set_pdeathsig,
            start_new_session=True,
        )

        # Start threads to monitor stdout/stderr
        def log_output(pipe, prefix):
            for line in pipe:
                if prefix == "stderr":
                    LOGGER.error(f"SSH {prefix}: {line.strip()}")
                else:
                    LOGGER.info(f"SSH {prefix}: {line.strip()}")

        threading.Thread(target=log_output, args=(ssh_process.stdout, "stdout"), daemon=True).start()
        threading.Thread(target=log_output, args=(ssh_process.stderr, "stderr"), daemon=True).start()

        # Accept the connection from the remote process with timeout
        LOGGER.debug("Waiting for remote connection")

        MAX_WAIT_TIME = 30
        local_socket.settimeout(2)
        start_time = time.time()
        while True:
            # Check if SSH process is still alive
            if ssh_process.poll() is not None:
                exit_code = ssh_process.poll()
                raise RuntimeError(f"SSH process died with exit code {exit_code} while waiting for connection")

            if time.time() - start_time > MAX_WAIT_TIME:
                raise RuntimeError("Timeout while waiting for remote connection")
            try:
                conn, _ = local_socket.accept()
                LOGGER.debug("Remote connection established")
                break
            except socket.timeout:
                LOGGER.debug("Still waiting for remote connection...")
                continue

        # Store the connection in the monitor instance, so that we can close it when cleanup is called
        monitor_instance.conn = conn
        monitor_instance.ssh_process = ssh_process
    except Exception as e:
        LOGGER.error(f"Error in setup_connection: {e}", exc_info=True)
        raise

    conn.settimeout(3)

    try:
        last_data: dict[str, datatype.Process] = {}
        while not shared_memory.is_finished.is_set():
            # Read message length (4 bytes)
            try:
                # LOGGER.debug("Reading message length")
                length_bytes = conn.recv(4)
            except socket.timeout:
                continue

            # LOGGER.debug("Received message length: %s", length_bytes)
            if not length_bytes:
                raise RuntimeError("Connection closed by remote")

            length = int.from_bytes(length_bytes, "big")
            # LOGGER.debug("Received message length: %d", length)

            # Read the full message
            data = conn.recv(length)

            try:
                info = json.loads(data.decode())
                if info.get("type") == "log":
                    # Handle log message
                    LOGGER.info("Remote: %s", info["message"])
                elif info.get("type") == "data":
                    # Handle process data
                    new_data = {
                        pid: datatype.Process(
                            pid=proc["pid"],
                            name=proc["name"],
                            cwd=proc["cwd"],
                            status=proc["status"],
                            create_time=proc["create_time"],
                            tcp=sorted(proc["tcp"]),
                            udp=sorted(proc["udp"]),
                        )
                        for pid, proc in info["processes"].items()
                    }
                    if new_data != last_data:
                        with shared_memory.lock:
                            LOGGER.debug("Setting new data")
                            shared_memory.processes = new_data
                            shared_memory.has_new_data.set()
                        last_data = new_data

            except json.JSONDecodeError as e:
                LOGGER.error("Error decoding JSON: %s", e)
                LOGGER.debug("Problematic data: %s", data)
    except socket.error as e:
        # socket is closing
        LOGGER.debug("Socket error in run_remote_script: %s", e)
        pass
    except Exception as e:
        LOGGER.debug("Exception in run_remote_script: %s", e, exc_info=True)
    finally:
        LOGGER.debug("Terminating SSH process")
        ssh_process.send_signal(signal.SIGINT)
        ssh_process.terminate()
        # os.killpg(os.getpgid(ssh_process.pid), signal.SIGTERM)

        try:
            ssh_process.wait(timeout=0.5)
        except subprocess.TimeoutExpired:
            LOGGER.warning("SSH process did not terminate gracefully, killing...")
            ssh_process.kill()
            ssh_process.wait()
        # LOGGER.debug("Closing local socket")
        # conn.close()
        # local_socket.close()
        # for thread in threads:
        #     thread.cancel()
        #     thread.join()
    LOGGER.debug("Remote socket script finished")


class RemoteProcessMonitor(AbstractProvider):
    def __init__(self, ssh_host: str):
        super().__init__()
        self.ssh_host = ssh_host
        LOGGER.debug("Initializing RemoteProcessMonitor for host: %s", ssh_host)
        self.shared_memory = SharedMemory(processes={})
        self.thread: threading.Thread | None = None
        self.cached_processes: dict[str, datatype.Process] = {}
        self.conn: socket.socket | None = None  # Store the socket connection
        self.ssh_process: subprocess.Popen | None = None
        self.forwarded_ports: dict[int, SSHForward] = {}

    @property
    def name(self) -> str:
        return f"{super().name}: {self.ssh_host}"

    def connect(self) -> bool:
        """Establish SSH connection and socket. Returns True if successful."""
        try:
            self.setup_connection()
            return True
        except Exception as e:
            LOGGER.error("Failed to establish connection: %s", e, exc_info=True)
            return False

    def setup_connection(self):
        self.thread = threading.Thread(target=run_remote_script, args=(self.ssh_host, self.shared_memory, self))
        self.thread.start()

    async def get_processes(self) -> dict[str, datatype.Process]:
        if self.shared_memory.has_new_data.is_set():
            with self.shared_memory.lock:
                self.cached_processes = self.shared_memory.processes
            self.shared_memory.has_new_data.clear()
        return self.cached_processes

    async def cleanup(self) -> None:
        # try to kill the ssh process to speed up cleanup
        if self.ssh_process:
            self.ssh_process.kill()

        LOGGER.debug("Cleaning up RemoteProcessMonitor")
        for port in self.forwarded_ports:
            self.forwarded_ports[port].cleanup()
        self.forwarded_ports.clear()

        with self.shared_memory.lock:
            self.shared_memory.is_finished.set()
        if self.conn:
            try:
                self.conn.shutdown(socket.SHUT_RDWR)
            except Exception as e:
                LOGGER.debug("Error during socket shutdown: %s", e)
            finally:
                self.conn.close()
                self.conn = None

        LOGGER.debug("Closing connection")
        if self.thread:
            self.thread.join()
            self.thread = None

    async def on_ports_turned_on(self, port: int):
        self.forwarded_ports[port] = SSHForward(port, ssh_host=self.ssh_host)
        try:
            # Start the reverse_port subprocess with process group
            self.forwarded_ports[port].start()
        except Exception as e:
            LOGGER.error("Failed to start port forwarding for port %s: %s", port, e)

    async def on_ports_turned_off(self, port: int):
        self.forwarded_ports[port].cleanup()
        self.forwarded_ports.pop(port)
